module Constants(Constant(..),Op(..),pConstant,pBool,pNumber,pOperator,Expr(..),
                 prec,arrow,arrows,int,contype,optype,pbracks_pcommas_e,decWith,assoc,pparse) where

import UU_Scanner
import UU_Parsing
import UU_Parsing_ext
import Char
import Numeric
import Types
import List
import Utils

data Constant = If | Number Int | FNumber Float | Binop Op | Boolean Bool | Nul | 
                Tuple Int | Constr String | Error
                                deriving (Eq, Ord, Show)
                                                    

pConstant = pNumber      		<|>
            Boolean <$> pBool           <|>
            Constr <$> pConid	

pBool = (const True) <$> pKey "True"	<|>
        (const False) <$> pKey "False"


pNumber = number <$> pInteger <*> pFraction
          where number n NoFrac = Number (stringToInt n)
                number n (Frac s) = FNumber (stringToFloat n + stringToFrac s)

data Fraction = Frac String | NoFrac

pFraction = (Frac <$-> pOper "." <*> pInteger) <|> succeed NoFrac 
stringToFloat :: String -> Float
stringToFloat = foldl step 0.0
                where step f c = f*10.0 + (fromInteger.toInteger) (ord c - ord '0')

stringToFrac :: String -> Float
stringToFrac = foldr step 0.0
               where step c f = ((fromInteger.toInteger) (ord c - ord '0') + f)/10.0


data Op =   App | Idx | Comp | 
            Exp | Tim | Div | Mod | Plus | Min | 
            FExp | FTim | FDiv | FPlus | FMin | 
            Cat | Cons | Eq  | Neq | Le  | Leq  | Ge  |
            Geq | And | Or  
				deriving (Eq, Ord, Show)

pOperator = const Idx  <$> pOper "!" <|>
            const Exp  <$> pOper "^" <|>
            const FExp <$> pOper "^." <|>
            const Comp <$> pOper "." <|>
            const Tim  <$> pOper "*" <|>
            const FTim <$> pOper "*." <|>
            const Div  <$> pOper "/" <|>
            const FDiv <$> pOper "/." <|>
            const Mod  <$> pKey "mod" <|>
            const Plus <$> pOper "+" <|>
            const FPlus <$> pOper "+." <|>
            const Min  <$> pOper "-" <|>
            const FMin <$> pOper "-." <|>
            const Cat  <$> pOper "++" <|>
            const Cons <$> pOper ":" <|>
            const Eq   <$> pOper "==" <|>
            const Neq  <$> pOper "/=" <|>
            const Le   <$> pOper "<" <|>
            const Leq  <$> pOper "<=" <|>
            const Ge   <$> pOper ">" <|>
            const Geq  <$> pOper ">=" <|>
            const And  <$> pOper "&&" <|>
            const Or   <$> pOper "||"

left Exp   = True
left FExp   = True
left Idx   = True
left Tim   = True
left FTim  = True
left Plus  = True
left FPlus = True
left Min   = True
left FMin  = True
left _     = False 

nonass App = True
nonass Div = True
nonass Mod = True
nonass Eq  = True
nonass Neq = True
nonass Le  = True
nonass Leq = True
nonass Ge  = True
nonass Geq = True
nonass _   = False

prec App = 100
prec Exp = 10 
prec FExp = 10 
prec Idx = 9
prec Comp = 9
prec Tim = 7
prec FTim = 7
prec Div = 7
prec FDiv = 7
prec Mod = 7
prec Plus = 6 
prec FPlus = 6 
prec Min = 6 
prec FMin = 6 
prec Cat = 5 
prec Cons = 5 
prec Eq = 4
prec Neq = 4
prec Le = 4
prec Leq = 4
prec Ge = 4
prec Geq = 4
prec And = 3
prec Or = 2

data Expr a b = Tip a | Node (Expr a b) b (Expr a b)

data Tree a = Nil | Bin (Tree a) a (Tree a)

decWith x t = fst (f t x)
              where f Nil x = (Tip (head x), tail x)
                    f (Bin s a t) x = (Node u a v,z)
                                      where (u,y) = f s x 
                                            (v,z) = f t y

assoc Nil = Nil
assoc e@(Bin e1 op e2) | nonass op = Bin (assoc e1) op (assoc e2)
                       | left op    = assl e 
                       | otherwise  = assr e 
                       

assl (Bin e1 op Nil) = Bin (assoc e1) op Nil
assl (Bin e1 op1 e@(Bin e2 op2 e3)) | op1 == op2 = assl (Bin (Bin e1 op1 e2) op2 e3)
                                    | otherwise = Bin (assoc e1) op1 (assoc e)

assr (Bin Nil op e1) = Bin Nil op (assoc e1)
assr (Bin e@(Bin e1 op1 e2) op2 e3) | op1 == op2 = assr (Bin e1 op1 (Bin e2 op2 e3))
                                    | otherwise = Bin (assoc e) op2 (assoc e3)

pparse = foldl bin Nil . foldr step []
         where step a x         = add a Nil x 
               add a t []       = [(a,t)]
               add a t ((b,s):x) | prec b <= prec a = (a,t):(b,s):x
                                 | otherwise        = add a (bin t (b,s)) x
               bin s (a,t)      = Bin s a t 


-- pitfall in UU_Parsing_ext: it does not allow an empty list of p's in a pListSep

pbracks_pcommas_e p = pbracks (succeed [] <|> pListSep (pSym ',') p)

arrow s t = TCon "Arrow" [s,t]
arrows xs = foldr1 arrow xs
bool = TCon "Bool" []
int = TCon "Int" []
float = TCon "Float" []
bint = arrows[ int,int,int]
bfloat = arrows [float,float,float]
bbool = arrows[ bool,bool,bool]
bcomp = close (arrows [a,a,bool])
tabs xs t = TAbs xs t
list t = TCon "List" [t]
var x = Tid x
a = var "a0"
b = var "b0"
c = var "c0"
d = var "d0"

contype env If = AType (close (arrows [bool, a, a, a]))
contype env (Number n) = AType int 
contype env (FNumber n) = AType float 
contype env (Binop op) = AType (optype op)
contype env (Boolean b) = AType bool 
contype env Nul = AType (close (list a))
contype env (Tuple n) = AType (close (arrows (as ++ [TCon ("Tuple"++show n) as])))
                        where as = map Tid (take n namesupply)
contype env (Constr x) = app env x
contype env Error = AType (close a)

optype Idx = close (arrows[ list a, int, a])
optype Comp = close (arrows[ arrow a b, arrow c a, arrow c b])
optype Exp = bint
optype Tim = bint 
optype Div = bint 
optype Mod = bint 
optype Plus = bint 
optype Min = bint 

optype FExp = bfloat
optype FTim = bfloat
optype FDiv = bfloat
optype FPlus = bfloat
optype FMin  = bfloat

optype Cat = close (arrows [ list a, list a, list a])
optype Cons = close (arrows [ a, list a, list a])
optype Eq = bcomp
optype Neq = bcomp
optype Le = bcomp
optype Leq = bcomp
optype Ge = bcomp
optype Geq = bcomp
optype And = bbool
optype Or = bbool

