-- Stand-in for SIMD floating-point 4-tuples.  Used for constant-folding.
-- We could use the real thing instead.  Should we?

module F32vec4 where

import PP
import Int
import NumExts

-- Uses Double for now, but really we should switch over to float.  We're
-- consistently using double at compile-time and float at display time.

-- N.B. Lower order floats come first.
type F32vec4 = (Double,Double,Double,Double)

-- The C++ constructor F32vec4 argument order is reversed from the memory
-- order.  We'll mimic both here, for sanity.

packF32 :: Double -> Double -> Double -> Double -> F32vec4
--packF32 = (,,,)
packF32 d c b a = (a,b,c,d)

nthF32 :: Int -> F32vec4 -> Double
-- nthF32 i (a,b,c,d) = [d,c,b,a]!!i
nthF32 i (a,b,c,d) = [a,b,c,d]!!i

sumF32 :: F32vec4 -> Double
sumF32 (a,b,c,d) = a + b + c + d

-- Unparse for C construction
ppF32vec4ForC :: F32vec4 -> PPDoc
ppF32vec4ForC (a,b,c,d) =
  text "F32vec4" <> ppTuple (map ppDouble [d,c,b,a])


instance Num F32vec4 where
  (a,b,c,d) + (a',b',c',d') = (a+a',b+b',c+c',d+d')
  (a,b,c,d) * (a',b',c',d') = (a*a',b*b',c*c',d*d')
  negate (a,b,c,d) = (negate a, negate b, negate c, negate d)
  fromInteger = doubleToF32vec4 . fromInteger
  fromInt = doubleToF32vec4 . fromInt

-- Hugs (at least) already has Ord for up to 5-tuples.  (Where?)
-- Moreover, it's inconsistent with what we want, since it chooses one
-- argument or the other, while simd fp makes four independent choices.
{-
instance Ord F32vec4 where
  _ <= _ = error "F32vec4: no <="
  (a,b,c,d) `min` (a',b',c',d') = (a `min` a',b `min` b',c `min` c',d `min` d')
  (a,b,c,d) `max` (a',b',c',d') = (a `max` a',b `max` b',c `max` c',d `max` d')
-}

(a,b,c,d) `minSimd` (a',b',c',d') = (a `min` a',b `min` b',c `min` c',d `min` d')
(a,b,c,d) `maxSimd` (a',b',c',d') = (a `max` a',b `max` b',c `max` c',d `max` d')

instance Fractional F32vec4 where
  recip (a,b,c,d) = (recip a, recip b, recip c, recip d)
  fromRational = error "F32vec4: no fromRational"
  -- don't use fromDouble: it's a Hugs only extension.

doubleToF32vec4 :: Double -> F32vec4
doubleToF32vec4 d = (d,d,d,d)

floatToF32vec4 :: Float -> F32vec4
floatToF32vec4 d = (f,f,f,f) where f = floatToDouble d

instance Floating F32vec4 where
  sqrt (a,b,c,d) = (sqrt a, sqrt b, sqrt c, sqrt d)
